@@ -1,9 +1,8 @@
 #coding: utf-8
 
-import os
-
-from pymorphy.constants import PRODUCTIVE_CLASSES, VERBS
+from pymorphy.constants import PRODUCTIVE_CLASSES, VERBS, NOUNS
 from pymorphy.constants import RU_CASES, RU_NUMBERS, RU_GENDERS, RU_PERSONS, RU_TENSES, RU_VOICES
+from pymorphy.constants import KEEP_GENDER_CLASSES, NORMAL_FORMS, NORMAL_FORMS_EN
 from pymorphy.backends import PickleDataSource, ShelveDataSource
 
 from utils import mprint
@@ -30,14 +29,16 @@ class GramForm(object):
     """ Класс для работы с грамматической формой """
 
     def __init__(self, form_string):
-        self.form = self.parse_str(form_string)
+        self.form, self.denied_form = self.parse_str(form_string)
 
     def parse_str(self, form_string):
-        attrs = [a for a in form_string.split(',') if a]
-        return set(attrs)
+        splitted = form_string.split(',')
+        form = [a for a in splitted if a and a[0]!='!']
+        denied_form = [a[1:] for a in splitted if a and a[0]=='!']
+        return set(form), set(denied_form)
 
     def get_form_string(self):
-        return ",".join(self.form)
+        return u",".join(self.form)
 
     def clear_number(self):
         ''' убрать информацию о числе '''
@@ -65,7 +66,7 @@ class GramForm(object):
 
     def update(self, form_string):
         """ Обновляет параметры, по возможности оставляя все, что можно. """
-        requested_form = self.parse_str(form_string)
+        requested_form = [a for a in form_string.split(',') if a and a[0]!='!']
 
         for item in requested_form:
 
@@ -96,6 +97,15 @@ class GramForm(object):
         self.form.update(requested_form)
         return self
 
+    def match(self, gram_form):
+        if not self.form.issuperset(gram_form.form): # не все параметры из gram_form есть тут
+            return False
+        if self.form.intersection(gram_form.denied_form):
+            return False
+        return True
+
+NORMAL_GRAM_FORMS = dict([(cls, GramForm(NORMAL_FORMS[cls][0]),) for cls in NORMAL_FORMS])
+
 
 class Morph:
     """ Класс, реализующий морфологический анализ на основе словарей из
@@ -106,18 +116,27 @@ class Morph:
                  predict_by_prefix = True, predict_by_suffix = True,
                  handle_EE = False):
         '''
-        data_source: источник данных. Наследник DictDataSource, свойства должны
-            поддерживать доступ по ключу.
+        Параметры конструктора:
 
-        check_prefixes: проверять ли вообще префиксы
-
-        predict_by_prefix: предсказывать ли по префиксу
-
-        predict_by_suffix: предсказывать ли по суффиксу
-
-        handle_EE: как обрабатывать букву ё. Если True, то все буквы ё считаются
-            равными е, если False - разными буквами. Значение должно совпадать
-            с тем, которое указано при конвертации словаря.
+        * data_source: источник данных. Может быть
+          :class:`ShelveDataSource`, :class:`MrdDataSource`,
+          :class:`PickleDataSource` или любой другой наследник
+          :class:`DictDataSource`, у которого есть атрибуты
+          rules, lemmas, prefixes, gramtab, rule_freq, endings,
+          possible_rule_prefixes, поддерживающие доступ по ключу.
+        * check_prefixes: проверять ли вообще префиксы
+        * predict_by_prefix: предсказывать ли по префиксу
+        * predict_by_suffix: предсказывать ли по суффиксу
+        * handle_EE: как обрабатывать букву ё. Если True, то все буквы ё
+          считаются равными е, если False - разными буквами. По умолчанию
+          в словарях все буквы ё заменены на е, и ожидается, что в
+          пользовательском вводе букв ё не будет (handle_EE=False).
+          Если словарь сконвертирован с handle_EE=True (в нем сохранены
+          буквы ё), то буквы е и ё будут считаться разными буквами, это
+          может привести к неожиданностям, т.к. часто вместо ё пишут е, а
+          в словаре будет слово только с ё. При handle_EE=True запросы
+          выполняются по 2 раза, со словом "как есть" и со словом, у
+          которого все е заменены на ё.
         '''
 
         self.data = data_source
@@ -132,10 +151,6 @@ class Morph:
 
         self.handle_EE = handle_EE
 
-    def normalize(self, word):
-        """ Вернуть нормальную форму слова """
-        return set([item['norm'] for item in self._get_graminfo(word)])
-
     def get_graminfo(self, word):
         """ Вернуть грамматическую информацию о слове """
         return self._get_graminfo(word)
@@ -144,6 +159,17 @@ class Morph:
         """
         Вернуть все варианты слова, соответствующие заданной
         грамматической форме и части речи.
+
+        Параметры:
+
+        * word - слово, которое хотим склонять
+        * gram_form - желаемые характеристики грам. формы. Выступает в роли
+          фильтра. Может быть пустым, тогда будут возвращены все
+          формы слова.
+        * gram_class - часть речи. Если передан этот параметр, то слово
+          будет считаться словом этой части речи (если возможно), и
+          склоняться будет соответственно. Если параметр не передан,
+          ограничения на часть речи накладываться не будут.
         """
         requested_attrs = gram_form.split(',')
         variants = []
@@ -160,11 +186,22 @@ class Morph:
         """
         Вернуть вариант слова, который соотвествует данной грамматической
         форме и части речи, а также менее всего отличается от исходного.
+
+        Параметры:
+
+        * word - слово, которое хотим склонять
+        * gram_form - желаемые характеристики грам. формы
+        * gram_class - часть речи. Если передан этот параметр, то слово
+          будет считаться словом этой части речи (если возможно), и
+          склоняться будет соответственно. Требуется для устранения
+          неоднозначностей.
         """
         forms = self.get_graminfo(word)
+        if gram_class is not None:
+            forms = [form for form in forms if form['class'] == gram_class]
         if not forms:
             return word
-        graminfo = forms[0]    #FIXME: первая форма в списке не обязательно самая лучшая
+        graminfo = forms[0]   #FIXME: первая форма в списке не обязательно самая лучшая
 
         form = GramForm(graminfo['info'])
         form.update(gram_form)
@@ -184,6 +221,76 @@ class Morph:
         return self.inflect_ru(word, form.get_form_string(), gram_class)
 
 
+    def pluralize_inflected_ru(self, word, num, gram_class=None):
+        """
+        Вернуть слово в форме, которая будет сочетаться с переданным числом.
+        Например: 1 попугай, 2 попугая, 5 попугаев.
+
+        Аналог choose_plural из pytils, для которого требуется только 1
+        начальная форма слова.
+        """
+        if num % 100 == 11: # заканчивается на 11
+            form = u"мн,рд"
+        elif num % 10 == 1: # заканчивается на 1
+            form = u"ед,им"
+        elif 11 < num % 100 < 20: # заканчивается на 12..19
+            form = u'мн,рд'
+        elif 2 <= num % 10 <= 4: # заканчивается на 2,3 или 4
+            form = u'ед,рд'
+        else:                   # заканчивается на 5..9 или 0
+            form = u'мн,рд'
+
+        return self.inflect_ru(word, form, gram_class)
+
+
+    def normalize(self, word):
+        """ Вернуть список нормальных форм слова """
+        forms = self.get_normal_forms(word)
+        if not forms:
+            return word
+        return set(form['word'] for form in forms)
+
+
+    def get_normal_forms(self, word):
+        """ Вернуть список нормальных форм слова с грам. информацией """
+        base_forms = self.get_graminfo(word)
+        if not base_forms:
+            return []
+
+        correct_genders = set()
+        for form in base_forms:
+            gram_form = GramForm(form['info']).form
+            for gender in RU_GENDERS:
+                if gender in gram_form:
+                    correct_genders.add(gender)
+
+        correct_classes = set([NORMAL_FORMS[form['class']][1] for form in base_forms])
+
+        # для англ. языка просто возвращаем первую форму
+        if correct_classes.issubset(set(NORMAL_FORMS_EN.keys())):
+            base_form = base_forms[0]
+            norm_form = base_form['lemma']+self.data.rules[base_form['paradigm_id']][0][0]
+            return [{'word': norm_form}]
+
+        variants = [variant for variant in self._decline(word) if variant['class'] in correct_classes]
+
+        def form_is_normal(form):
+            gram_form = GramForm(form['info'])
+            normal_form = NORMAL_GRAM_FORMS[form['class']]
+            if form['class'] in KEEP_GENDER_CLASSES:
+                if not set(gram_form.form).intersection(correct_genders):
+                    return False
+            return gram_form.match(normal_form)
+
+
+        if correct_genders:
+            normal_forms = [form for form in variants if form_is_normal(form)]
+        else:
+            normal_forms = [form for form in variants if GramForm(form['info']).match(NORMAL_GRAM_FORMS[form['class']])]
+
+        return normal_forms
+
+
 #----------- internal methods -------------
 
     def _drop_cache(self):
@@ -197,7 +304,6 @@ class Morph:
         """ Просклонять: вернуть все грам. формы с информацией про них """
 
         word_graminfo = self.get_graminfo(src_word)
-
         forms = []
 
         # убираем дубликаты парадигм
@@ -227,6 +333,7 @@ class Morph:
                 })
         return forms
 
+
     def _get_lemma_graminfo(self, lemma, suffix, require_prefix, method_format_str):
         """ Получить грам. информацию по лемме и суффиксу. Для леммы перебираем все
             правила, смотрим, есть ли среди них такие, которые приводят к
@@ -238,22 +345,24 @@ class Morph:
         # для леммы смотрим все доступные парадигмы
         for paradigm_id in lemma_paradigms:
             paradigm = data_source.rules[paradigm_id]
-            # все правила в парадигме
-            for rule in paradigm:
-                rule_suffix, rule_ancode, rule_prefix = rule
-                # если по правилу выходит, что окончание такое, как надо,
-                # то значит нашли, что искали
-                if rule_suffix==suffix and rule_prefix==require_prefix:
-                    graminfo = data_source.gramtab[rule_ancode]
-                    norm_form = lemma + paradigm[0][0]   #FIXME: первая по порядку форма не обязательно нормальная!
-                    gram.append({'norm': norm_form,
-                                 'class': graminfo[0],
-                                 'info': graminfo[1],
-                                 'paradigm_id': paradigm_id,
-                                 'ancode': rule_ancode,
-                                 'lemma': lemma,
-                                 'method': method_format_str % (lemma, suffix)
-                               })
+            # оставляем только те правила, по которым можно слово составить
+            valid_rules = [rule for rule in paradigm if rule[0]==suffix and rule[2]==require_prefix]
+            if not valid_rules:
+                continue
+            for rule in valid_rules:
+                ancode = rule[1]
+                graminfo = data_source.gramtab[ancode]
+                gram_form = {
+                         'class': graminfo[0],
+                         'info': graminfo[1],
+                         'paradigm_id': paradigm_id,
+                         'ancode': ancode,
+                         'lemma': lemma,
+                         'method': method_format_str % (lemma, suffix)
+                        }
+                # не допускаем дубликатов
+                if not gram_form in gram:
+                    gram.append(gram_form)
         return gram
 
 
@@ -301,8 +410,7 @@ class Morph:
                             # на суффикс начальной формы
                             suffix_len = len(suffix)
                             predicted_lemma = word[0:-suffix_len] if suffix_len else word
-                            norm_form = predicted_lemma + rules_list[0][0]
-                            gram.append({'norm': norm_form,
+                            gram.append({
                                          'class':graminfo[0],
                                          'info': graminfo[1],
                                          'paradigm_id': paradigm_id,
@@ -339,7 +447,6 @@ class Morph:
 
                 # приписываем префикс обратно к полученным нормальным формам
                 for form in base_forms:
-                    form['norm'] = prefix+form['norm']
                     form['prefix'] = prefix
                     form['method'] = 'prefix(%s).%s' % (prefix, form['method'])
                 gram.extend(base_forms)
@@ -385,7 +492,6 @@ class Morph:
 
             # приписываем префикс обратно
             for form in base_forms:
-                form['norm'] = prefix+form['norm']
                 form['predict-prefix'] = prefix
                 form['method'] = 'predict-prefix(%s).%s' % (prefix, form['method'])
             gram.extend(base_forms)
@@ -403,7 +509,6 @@ class Morph:
         gram = self._get_graminfo(word.replace(u'Е', u'Ё'), require_prefix,
                                   predict_EE = False)
         for info in gram:
-            info['norm'] = info['norm'].replace(u'Ё', u'Е')
             info['method'] = info['method'].replace(u'Ё', u'Е')
         return gram
 
@@ -458,9 +563,9 @@ def get_morph(dir, backend='shelve', cached=True, **kwargs):
 
     Параметры:
 
-        * dir - путь к папке с файлами словарей
-        * backend - тип словарей. Может быть 'shelve', 'tch', 'tcb', 'cdb'.
-        * cached - кешировать ли данные в оперативной памяти
+    * dir - путь к папке с файлами словарей
+    * backend - тип словарей. Может быть 'shelve', 'tch', 'tcb', 'cdb'.
+    * cached - кешировать ли данные в оперативной памяти
 
     Также можно указывать все параметры, которые принимает конструктор класса
     Morph.
@@ -470,6 +575,12 @@ def get_morph(dir, backend='shelve', cached=True, **kwargs):
 
 
 def get_pickle_morph(filename, **kwargs):
+    """
+    Вернуть объект с морфологическим анализатором (Morph), используя
+    pickle - хранилище в качестве источника данных. Операции будут выполняться
+    быстрее, но память расходуется очень неэффективно (потребуется порядка
+    200 Мб). Так что использовать очень осторожно.
+    """
     return Morph(PickleDataSource(filename), **kwargs)
 
 
@@ -480,6 +591,8 @@ def setup_psyco():
         from pymorphy.backends.shelve_source.shelf_with_hooks import ShelfWithHooks
         psyco.bind(Morph._get_graminfo)
         psyco.bind(Morph._get_lemma_graminfo)
+        psyco.bind(Morph.get_normal_forms)
+        psyco.bind(GramForm.__init__)
         psyco.bind(ShelfWithHooks._getitem__cached)
         psyco.bind(ShelfWithHooks._contains__cached)
         psyco.bind(_get_split_variants)
